/*-
 * Copyright 2013 - 2018 UPLEX Nils Goroll Systemoptimierung
 * All rights reserved
 *
 * Authors: Geoffrey Simmons <geoffrey.simmons@uplex.de>
 *	    Nils Goroll <nils.goroll@uplex.de>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include "config.h"

#include <stdlib.h>
#include <string.h>

#include "cache/cache.h"
#include "vcl.h"
#include "vre.h"
#include "vas.h"
#include "vsb.h"
#include "miniobj.h"

#include "vcc_if.h"

/* pcreapi(3):
 * 
 * The first two-thirds of the vector is used to pass back captured
 * substrings, each substring using a pair of integers. The remaining
 * third of the vector is used as workspace by pcre_exec() while matching
 * capturing subpatterns, and is not available for passing back
 * information.
 *
 * XXX: if vre were to expose the pcre and pcre_extra objects, then we
 * could use pcre_fullinfo() to determine the highest backref for each
 * regex, and wouldn't need this arbitrary limit ...
 */
#define MAX_MATCHES	11
#define MAX_OV		((MAX_MATCHES) * 3)
#define MAX_OV_USED	((MAX_MATCHES) * 2)

struct vmod_re_regex {
	unsigned		magic;
#define VMOD_RE_REGEX_MAGIC 0x955706ee
	vre_t			*vre;
	struct vre_limits	vre_limits;
};

typedef struct ov_s {
	unsigned		magic;
#define OV_MAGIC 0x844bfa39
	const char		*subject;
	int			ovector[MAX_OV_USED];
} ov_t;

static void
errmsg(VRT_CTX, const char *fmt, ...)
{
	va_list args;

	AZ(ctx->method & VCL_MET_TASK_H);
	va_start(args, fmt);
	if (ctx->vsl)
		VSLbv(ctx->vsl, SLT_VCL_Error, fmt, args);
	else
		VSLv(SLT_VCL_Error, 0, fmt, args);
	va_end(args);
}

VCL_VOID
vmod_regex__init(VRT_CTX, struct vmod_re_regex **rep, const char *vcl_name,
		 VCL_STRING pattern, VCL_INT limit, VCL_INT limit_recursion)
{
	struct vmod_re_regex *re;
	vre_t *vre;
	int erroffset;
	const char *error;

	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
	AN(rep);
	AZ(*rep);
	AN(vcl_name);
	AN(pattern);

	if (limit < 1) {
		VRT_fail(ctx, "vmod re: invalid limit %d in %s constructor",
			 limit, vcl_name);
		return;
	}

	if (limit_recursion < 1) {
		VRT_fail(ctx, "vmod re: invalid limit_recursion %d "
			 "in %s constructor", limit_recursion, vcl_name);
		return;
	}

	if ((vre = VRE_compile(pattern, 0, &error, &erroffset)) == NULL) {
		VRT_fail(ctx, "vmod re: error compiling regex \"%s\" in %s "
			 "constructor: %s (at offset %d)", pattern, vcl_name,
			 error, erroffset);
		return;
	}

	ALLOC_OBJ(re, VMOD_RE_REGEX_MAGIC);
	AN(re);
	re->vre = vre;
	re->vre_limits.match = limit;
	re->vre_limits.match_recursion = limit_recursion;
	*rep = re;
}

VCL_VOID
vmod_regex__fini(struct vmod_re_regex **rep)
{
	struct vmod_re_regex *re;

	if (rep == NULL || *rep == NULL)
		return;
	re = *rep;
	*rep = NULL;
	CHECK_OBJ_NOTNULL(re, VMOD_RE_REGEX_MAGIC);
	if (re->vre != NULL)
		VRE_free(&re->vre);
	FREE_OBJ(re);
}

static VCL_BOOL
match(VRT_CTX, vre_t *vre, VCL_STRING subject, struct vmod_priv *task,
	const struct vre_limits *vre_limits)
{
	ov_t *ov;
	int s, nov[MAX_OV];
	size_t cp, len;

	AN(vre);
	if (subject == NULL)
		subject = "";

	if (task->priv == NULL) {
		if ((task->priv = WS_Alloc(ctx->ws, sizeof(*ov))) == NULL) {
			VSLb(ctx->vsl, SLT_VCL_Error, "vmod re error: "
			     "insufficient workspace for backref data");
			return 0;
		}
		task->len = -1;
		task->free = NULL;
		ov = (ov_t *) task->priv;
		ov->magic = OV_MAGIC;
	}
	else {
		WS_Assert_Allocated(ctx->ws, task->priv, sizeof(*ov));
		CAST_OBJ(ov, task->priv, OV_MAGIC);
	}

	len = strlen(subject);
	s = VRE_exec(vre, subject, len, 0, 0, nov, MAX_OV, vre_limits);
	if (s <= VRE_ERROR_NOMATCH) {
		if (s < VRE_ERROR_NOMATCH)
			VSLb(ctx->vsl, SLT_VCL_Error,
			     "vmod re: regex match returned %d", s);
		return 0;
	}
	if (s == 0) {
		VSLb(ctx->vsl, SLT_VCL_Error,
	     	     "vmod re: capturing substrings exceed max %d",
		     MAX_MATCHES - 1);
		s = MAX_MATCHES;
	}

	task->len = sizeof(*ov);
	if (WS_Inside(ctx->ws, subject, subject + len))
		ov->subject = subject;
	else if ((ov->subject = WS_Copy(ctx->ws, (const void *) subject, len))
		 == NULL) {
		VSLb(ctx->vsl, SLT_VCL_Error,
		     "vmod re: insufficient workspace");
		return 0;
	}
	memset(ov->ovector, -1, sizeof(ov->ovector));
	cp = s * 2 * sizeof(*nov);
	assert(cp <= sizeof(nov));
	memcpy(ov->ovector, nov, cp);
	return 1;
}

static VCL_STRING
backref(VRT_CTX, VCL_INT refnum, VCL_STRING fallback, struct vmod_priv *task)
{
	ov_t *ov;
	char *substr;
	int len;

	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
	AN(fallback);
	AN(task);

	if (refnum < 0 || refnum >= MAX_MATCHES) {
		VSLb(ctx->vsl, SLT_VCL_Error,
		     "vmod re: backref %ld out of range", refnum);
		return fallback;
	}

	if (task->priv == NULL) {
		VSLb(ctx->vsl, SLT_VCL_Error,
		     "vmod re: backref called without prior match");
		return fallback;
	}
	if (task->len <= 0)
		return fallback;

	WS_Assert_Allocated(ctx->ws, task->priv, sizeof(*ov));
	CAST_OBJ(ov, task->priv, OV_MAGIC);

	refnum <<= 1;
	assert(refnum + 1 < MAX_OV_USED);
	if (ov->ovector[refnum] == -1)
		return fallback;

	len = ov->ovector[refnum+1] - ov->ovector[refnum];
	assert(len <= ov->ovector[1] - ov->ovector[0]);
	WS_Assert_Allocated(ctx->ws, ov->subject + ov->ovector[refnum], len);
	substr = WS_Printf(ctx->ws, "%.*s", len,
			   ov->subject + ov->ovector[refnum]);
	if (substr == NULL) {
		VSLb(ctx->vsl, SLT_VCL_Error,
		     "vmod re: insufficient workspace");
		return fallback;
	}
	return substr;
}

static inline const struct vre_limits *
get_limits(const struct vmod_re_regex *re, struct vre_limits *limits,
	   VCL_INT limit, VCL_INT limit_recursion)
{
	if (limit <= 0 && limit_recursion <= 0)
		return (&re->vre_limits);

	if (limit > 0)
		limits->match = limit;
	else
		limits->match = re->vre_limits.match;

	if (limit_recursion > 0)
		limits->match_recursion = limit_recursion;
	else
		limits->match_recursion = re->vre_limits.match_recursion;

	return (limits);
}

VCL_BOOL
vmod_regex_match(VRT_CTX, struct vmod_re_regex *re, VCL_STRING subject,
		 VCL_INT limit, VCL_INT limit_recursion)
{
	struct vmod_priv *task;
	struct vre_limits buf;

	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
	CHECK_OBJ_NOTNULL(re, VMOD_RE_REGEX_MAGIC);
	AN(re->vre);

	task = VRT_priv_task(ctx, re);

	if (task == NULL) {
		errmsg(ctx, "vmod re: no priv - out of workspace?");
		return (0);
	}

	AN(task);
	task->len = 0;

	return match(ctx, re->vre, subject, task,
		     get_limits(re, &buf, limit, limit_recursion));
}

VCL_STRING
vmod_regex_backref(VRT_CTX, struct vmod_re_regex *re, VCL_INT refnum,
		   VCL_STRING fallback)
{
	struct vmod_priv *task;

	CHECK_OBJ_NOTNULL(re, VMOD_RE_REGEX_MAGIC);

	task = VRT_priv_task(ctx, re);
	if (task == NULL) {
		errmsg(ctx, "vmod re: no priv - out of workspace?");
		return (0);
	}
	return backref(ctx, refnum, fallback, task);
}

VCL_BOOL
vmod_match_dyn(VRT_CTX, struct vmod_priv *task, VCL_STRING pattern,
	       VCL_STRING subject, VCL_INT limit, VCL_INT limit_recursion)
{
	vre_t *vre;
	int erroffset;
	const char *error;
	VCL_BOOL dyn_return;
	struct vre_limits limits;

	AN(pattern);
	AN(task);

	if (limit < 1) {
		errmsg(ctx, "vmod re: invalid limit %d for regex \"%s\"",
		       limit, pattern);
		return 0;
	}

	if (limit_recursion < 1) {
		errmsg(ctx, "vmod re: invalid limit_recursion %d "
		       "for regex \"%s\"", limit_recursion, pattern);
		return 0;
	}

	limits.match = limit;
	limits.match_recursion = limit_recursion;

	task->len = 0;
	vre = VRE_compile(pattern, 0, &error, &erroffset);
	if (vre == NULL) {
		VSLb(ctx->vsl, SLT_VCL_Error,
		     "vmod re: error compiling regex \"%s\": %s (position %d)",
		     pattern, error, erroffset);
		return 0;
	}

	dyn_return = match(ctx, vre, subject, task, &limits);

	VRE_free(&vre);
	return dyn_return;
}

VCL_STRING
vmod_backref_dyn(VRT_CTX, struct vmod_priv *task, VCL_INT refnum,
		 VCL_STRING fallback)
{
	return backref(ctx, refnum, fallback, task);
}

VCL_STRING
vmod_version(VRT_CTX __attribute__((unused)))
{
	return VERSION;
}
